# Notification
serviceNotifyNumbers = []
serviceNotifyFrom = ""
spaNotifyNumbers = []
spaNotifyFrom = ""
doorbellNotifyNumbers = []
doorbellNotifyFrom = ""
doorNotifyNumbers = []
doorNotifyFrom = ""

import requests
import urllib.request, urllib.parse, urllib.error
import json
from twilio.rest import Client
from homealone import *

twilioKey = keyDir+"twilio.key"
iftttKey = keyDir+"ifttt.key"

# send notifications
def notify(resources, notificationType, message):
    debug("debugNotification", "notification", notificationType, message)
    if resources.getRes(notificationType).getState():
        if resources.getRes("smsAlerts").getState():
            if notificationType == "alertServices":
                fromNumber = serviceNotifyFrom
                toNumbers = serviceNotifyNumbers
            elif notificationType == "alertSpa":
                fromNumber = spaNotifyFrom
                toNumbers = spaNotifyNumbers
            elif notificationType == "alertDoorbell":
                fromNumber = doorbellNotifyFrom
                toNumbers = doorbellNotifyNumbers
            elif notificationType == "alertDoors":
                fromNumber = doorNotifyFrom
                toNumbers = doorNotifyNumbers
            elif notificationType == "alertMotion":
                fromNumber = motionNotifyFrom
                toNumbers = motionNotifyNumbers
            smsNotify(fromNumber, toNumbers, message)
        if resources.getRes("appAlerts").getState():
            appNotify("", message)
        if resources.getRes("iftttAlerts").getState():
            iftttNotify(message)
    else:
        debug("debugNotification", "notification", notificationType, "not enabled")

# send an sms notification
def smsNotify(fromNumber, toNumbers, message):
    smsClient = Client(getValue(twilioKey, "sid"), getValue(twilioKey, "token"))
    for toNumber in toNumbers:
        debug("debugNotification", "SMS notify from", fromNumber, "to", toNumber)
        smsClient.messages.create(to=toNumber, from_=fromNumber, body=message)

# send an app notification
def appNotify(app, message):
    if app != "":
        debug("debugNotification", "app notify to", app)
        requests.get("http://"+app+".appspot.com/notify?message="+urllib.parse.quote(message))

# send an IFTTT notification
def iftttNotify(message):
    key = getValue(iftttKey, "key")
    debug("debugNotification", "IFTTT notify")
    url = "https://maker.ifttt.com/trigger/haEvent/with/key/"+key
    headers = {"Content-Type": "application/json"}
    value2 = ""
    value3 = ""
    data = json.dumps({"value1": message, "value2": value2, "value3": value3})
    req = requests.post(url, data=data, headers=headers)



# application template

from .core import *
from .schedule import *
from .rest.restServer import *
from .rest.restProxy import *
from .metrics.metrics import *
from .interfaces.fileInterface import *

class Application(object):
    def __init__(self, name, globals,
                 publish=True, port=restServicePort,
                 state=False, shared=False, changeMonitor=True,
                 metrics=False, # sendMetrics=False, logMetrics=True, backupMetrics=True, purgeMetrics=False, purgeDays=5, logChanged=True,
                 proxy=False, watch=[], ignore=[]):
        self.name = name
        self.globals = globals                      # application global variables
        self.publish = publish                      # run REST server if true
        self.port = port                            # REST server port
        self.state = state
        self.shared = shared
        self.changeMonitor = changeMonitor
        self.stateInterface = None                  # Interface resource for state file
        self.metrics = metrics                      # publish metrics if true
        # self.sendMetrics = sendMetrics
        # self.logMetrics = logMetrics
        # self.backupMetrics = backupMetrics
        # self.purgeMetrics = purgeMetrics
        # self.purgeDays = purgeDays
        # self.logChanged = logChanged
        self.event = threading.Event()              # state change event
        self.resources = Collection("resources", event=self.event)    # resources to be published by REST server
        self.schedule = Schedule("schedule")        # schedule of tasks to run
        self.startList = []                         # resources that need to be started
        self.proxy = proxy
        self.restProxy = None
        self.proxyResources = None

        if state:
            if not os.path.exists(stateDir):
                os.mkdir(stateDir)
            self.stateInterface = FileInterface("stateInterface", fileName=stateDir+self.name+".state", shared=shared, changeMonitor=changeMonitor)
            self.stateInterface.start()
            self.globals["stateInterface"] = self.stateInterface
        if proxy:
            self.proxyResources = Collection("proxyResources", event=self.event)
            self.restProxy = RestProxy("restProxy", self.proxyResources, watch=watch, ignore=ignore, event=self.event)

    # start the application processes
    def run(self):
        if self.proxy:
            self.restProxy.start()
        if self.metrics:
            startMetrics(self.resources)
                # self.sendMetrics, self.logMetrics, self.backupMetrics, self.purgeMetrics, self.purgeDays, self.logChanged)
        for resource in self.startList:
            resource.start()
        if list(self.schedule.keys()) != []:
            self.schedule.start()
        if self.publish:
            self.restServer = RestServer(self.name, self.resources, port=self.port, event=self.event, label=self.name)
            self.restServer.start()

    # define an Interface resource
    def interface(self, interface, event=False, start=False):
        self.globals[interface.name] = interface
        if event:
            interface.event = self.event
        if start:
            self.startList.append(interface)

    # define a Sensor or Control resource
    def resource(self, resource, id=None, event=False, publish=True, start=False):
        if not id:
            id = camelize(resource.name)
        self.globals[id] = resource
        if event:
            resource.event = self.event
        if publish:
            self.resources.addRes(resource)
        if start:
            self.startList.append(resource)

    # define a Sensor or Control resource that is proxied from another server
    def proxyResource(self, resource, id=None):
        if not id:
            id = camelize(resource.name)
        self.globals[id] = resource
        resource.resources = self.proxyResources

    # define a Task resource
    def task(self, task, event=True, publish=True):
        self.schedule.addTask(task)
        self.globals[task.name] = self.schedule[task.name]
        if event:
            task.event = self.event
        if publish:
            self.resources.addRes(task)

    # apply a UI style to one or more resources
    def style(self, style, resources=[]):
        if resources == []:     # default is all resources
            resources = list(self.resources.values())
        for resource in listize(resources):
            resource.type = style

    # associate one or more resources with one or more UI groups
    def group(self, group, resources=[]):
        if resources == []:     # default is all resources
            resources = list(self.resources.values())
        for resource in listize(resources):
            resource.group = group

    # add a UI label to one or more resources
    def label(self, label=None, resources=[], strip=""):
        if resources == []:     # default is all resources
            resources = list(self.resources.values())
        for resource in listize(resources):
            if label:
                resource.label = label
            else:               # create a label from the name
                resource.label = labelize(resource.name.strip(strip))




import requests
import urllib.request, urllib.parse, urllib.error
import subprocess
from homealone import *

def notify(notificationType, message):
    servers = subprocess.check_output("avahi-browse -tp --resolve _notification._tcp" ,shell=True).decode().split("\n")
    if servers == [""]:
        log("notificationClient", "server not found")
    else:
        for server in servers:
            serverData = server.split(";")
            if len(serverData) > 6:
                host = serverData[6]
                port = serverData[8]
                url = "http://"+host+":"+port+"/notify?eventType="+notificationType+"&message="+urllib.parse.quote(message)
                request = requests.get(url)
                if request.status_code != 200:
                    log("notificationClient", "error", url, request.status_code)
                break



# initialize the global variables that define the environment
import os
import socket
import sys

# network
hostname = socket.gethostname()
localController = "localhost"

# directory structure
rootDir = sys.path[0]+"/"        # directory containing this application
keyDir = rootDir+"keys/"
stateDir = rootDir+"state/"
soundDir = rootDir+"sounds/"
dataDir = rootDir+"data/"

# logging and debugging
sysLogging = True
debugEnable = False

# Localization
latLong = (0.0, 0.0)
elevation = 0 # elevation in feet
tempScale = "F"

# Metrics
sendMetrics=False
logMetrics=True
backupMetrics=True
purgeMetrics=False
purgeDays=5
logChanged=True

# optionally import local variables
try:
    from conf import *
except ImportError:
    pass



tcpTimeout = 10.0
pollInterval = 1
disabledPollInterval = 10
maxRetries = 3

import socket
import json
import struct
import threading
import time
from homealone import *

# TP-Link device control

port = 9999

# Encryption and Decryption of TP-Link Smart Home Protocol
# XOR Autokey Cipher with starting key = 171
def encrypt(string):
    msg = bytes(string, "utf-8")
    result = struct.pack('>I', len(msg))
    key = 171
    for i in msg:
        a = key ^ int(i)
        key = a
        result += bytes([a])
    return result

def decrypt(msg):
    result = b""
    key = 171
    for i in msg[4:]:
        a = key ^ int(i)
        key = i
        result += bytes([a])
    return result.decode("utf-8")

class TplinkInterface(Interface):
    def __init__(self, name, ipAddr, model=None, interface=None, event=None):
        Interface.__init__(self, name, interface, event=event)
        self.ipAddr = ipAddr            # IP address of TPLink device
        self.model = model              # TPLink model number
        self.sysInfo = {}               # TPLink device information
        self.sleepTime = pollInterval   # how often to poll for state data
        self.errorCount = 0             # number of consecutive read errors
        # poll the device every second to generate state change notifications
        # cached state is the dictionary that is returned
        def getInfo():
            debug("debugTplink", self.name, "getInfo starting")
            while True:
                try:
                    sysInfo = self.readSysInfo()
                    if sysInfo:
                        debug("debugTplink", self.name, "sysInfo:", sysInfo)
                        self.sysInfo = sysInfo
                        if not self.model:
                            self.model = sysInfo["model"][0:5]
                        for sensor in list(self.sensors.values()):
                            if isinstance(sensor.addr, int):    # sensor is the device that is controlled
                                if self.model in ["KP200"]:      # TPLink device controls multiple devices
                                    sensorState = self.sysInfo["children"][sensor.addr]["state"]
                                else:                           # TPLink device controls a single device
                                    sensorState = self.sysInfo["relay_state"]
                            else:                               # sensor is some other attribute of the TPLink device
                                sensorState = self.sysInfo[sensor.addr]
                            if sensorState != self.states[sensor.addr]: # state has changed
                                self.states[sensor.addr] = sensorState
                                sensor.notify()
                except Exception as ex:
                    log("tplink state exception", self.sensorAddrs[ipAddr].name, ipAddr)
                    logException(self.name, ex)
                time.sleep(self.sleepTime)
            debug("debugTplink", self.name, "getInfo terminated")
        stateThread = LogThread(name=self.name+" getInfo", target=getInfo)
        stateThread.start()

    def readSysInfo(self):
        debug("debugTplink", self.name, "readSysInfo", self.ipAddr)
        try:
            tcpSocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            tcpSocket.connect((self.ipAddr, port))
            tcpSocket.settimeout(tcpTimeout)
            tcpSocket.send(encrypt('{"system":{"get_sysinfo":{}}}'))
            data = tcpSocket.recv(2048)
            tcpSocket.close()
            if not self.enabled:
                log(self.name, "enabling after", self.errorCount, "retries")
                self.enable()
                self.sleepTime = pollInterval
                self.errorCount = 0
            sysInfo = json.loads(decrypt(data))["system"]["get_sysinfo"]
            return sysInfo
        except Exception as ex:
            self.errorCount += 1
            if self.enabled:
                log(self.name, "read exception", str(ex), self.ipAddr)
                if self.errorCount == maxRetries:
                    log(self.name, "disabling after", maxRetries, "retries")
                    self.disable()
                    self.sleepTime = disabledPollInterval
            return None

    def read(self, addr):
        debug("debugTplink", self.name, "read", addr)
        try:
            return self.states[addr]
        except Exception as ex:
            logException(self.name, ex)
            debug("debugTplink", self.name, "state", self.sysInfo)
            return None

    def write(self, addr, state):
        debug("debugTplink", self.name, "write", addr, state)
        if not self.enabled:
            return None
        try:
            # only relay_state can be written
            if isinstance(addr, int):
                if self.model in ["KP200"]:      # TPLink device controls multiple devices
                    deviceContext = ',"context":{"child_ids":["'+self.sysInfo["deviceId"]+'%02d'%addr+'"]}'
                else:
                    deviceContext = ''
                stateMsg = '{"system":{"set_relay_state":{"state":'+str(state)+'}}'+deviceContext+'}'
                debug("debugTplink", self.name, "stateMsg:", stateMsg)
                tcpSocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                tcpSocket.connect((self.ipAddr, port))
                tcpSocket.settimeout(tcpTimeout)
                tcpSocket.send(encrypt(stateMsg))
                statusMsg = decrypt(tcpSocket.recv(2048))
                debug("debugTplink", self.name, "statusMsg:", statusMsg)
                status = int(json.loads(statusMsg)["system"]["set_relay_state"]["err_code"])
                tcpSocket.close()
                if status == 0:
                    # update the cached state
                    self.states[addr] = state
                    self.sensorAddrs[addr].notify()
                    return state
                else:
                    return None
            else:
                return None
        except Exception as ex:
            log("tplink write exception", self.sensorAddrs[ipAddr].name, ipAddr)
            logException(self.name, ex)
            return None
